import { LabIcon } from '@jupyterlab/ui-components';
import { each } from '@lumino/algorithm';
import { CellToolbarWidget } from './celltoolbarwidget';
import { EXTENSION_ID } from './tokens';
// icon svg import statements
import addAboveSvg from '../style/icons/addabove.svg';
import addBelowSvg from '../style/icons/addbelow.svg';
import deleteSvg from '../style/icons/delete.svg';
import duplicateSvg from '../style/icons/duplicate.svg';
import moveDownSvg from '../style/icons/movedown.svg';
import moveUpSvg from '../style/icons/moveup.svg';
const DEFAULT_LEFT_MENU = [];
/**
 * Widget cell toolbar class
 */
const CELL_BAR_CLASS = 'jp-cell-bar';
/**
 * Class for a cell whose contents overlap with the cell toolbar
 */
const TOOLBAR_OVERLAP_CLASS = 'jp-toolbar-overlap';
/**
 * Icons for use in toolbar.
 *
 * These are copied from icon.ts, which is not part of the webpack bundle
 * because nothing is imported from it.
 */
export const addAboveIcon = new LabIcon({
    name: `${EXTENSION_ID}:add-above`,
    svgstr: addAboveSvg
});
export const addBelowIcon = new LabIcon({
    name: `${EXTENSION_ID}:add-below`,
    svgstr: addBelowSvg
});
export const deleteIcon = new LabIcon({
    name: `${EXTENSION_ID}:delete`,
    svgstr: deleteSvg
});
export const duplicateIcon = new LabIcon({
    name: `${EXTENSION_ID}:duplicate`,
    svgstr: duplicateSvg
});
export const moveDownIcon = new LabIcon({
    name: `${EXTENSION_ID}:move-down`,
    svgstr: moveDownSvg
});
export const moveUpIcon = new LabIcon({
    name: `${EXTENSION_ID}:move-up`,
    svgstr: moveUpSvg
});
/**
 * Watch a notebook so that a cell toolbar appears on the active cell
 */
export class CellToolbarTracker {
    constructor(panel, commands, settings) {
        this._isDisposed = false;
        this._commands = commands;
        this._panel = panel;
        this._settings = settings;
        this._previousActiveCell = this._panel.content.activeCell;
        if (this._settings) {
            this._onSettingsChanged();
            this._settings.changed.connect(this._onSettingsChanged, this);
        }
        const notebookModel = this._panel.context.model;
        const cells = notebookModel.cells;
        cells.changed.connect(this.updateConnectedCells, this);
        // Only add the toolbar to the notebook's active cell (if any) once it has fully rendered and been revealed.
        panel.revealed.then(() => this._onActiveCellChanged(panel.content));
        // Handle subsequent changes of active cell.
        panel.content.activeCellChanged.connect(this._onActiveCellChanged, this);
    }
    _onActiveCellChanged(notebook) {
        const activeCell = notebook.activeCell;
        if (!activeCell) {
            return;
        }
        if (this._previousActiveCell !== null &&
            this._previousActiveCell !== undefined) {
            this._removeToolbar(this._previousActiveCell.model);
        }
        this._addToolbar(activeCell.model);
        this._previousActiveCell = activeCell;
        this._updateCellForToolbarOverlap(activeCell);
    }
    get isDisposed() {
        return this._isDisposed;
    }
    dispose() {
        var _a;
        if (this.isDisposed) {
            return;
        }
        this._isDisposed = true;
        if (this._settings) {
            this._settings.changed.disconnect(this._onSettingsChanged, this);
        }
        const cells = (_a = this._panel) === null || _a === void 0 ? void 0 : _a.context.model.cells;
        if (cells) {
            cells.changed.disconnect(this.updateConnectedCells, this);
            each(cells.iter(), model => this._removeToolbar(model));
        }
        this._panel = null;
    }
    /**
     * Callback to react to cells list changes
     *
     * @param cells List of notebook cells
     * @param changed Modification of the list
     */
    updateConnectedCells(cells, changed) {
        var _a;
        const activeCell = (_a = this._panel) === null || _a === void 0 ? void 0 : _a.content.activeCell;
        if (activeCell === null || activeCell === undefined) {
            return;
        }
        if (changed.oldValues.find(m => m === activeCell.model)) {
            this._removeToolbar(activeCell.model);
            this._addToolbar(activeCell.model);
        }
    }
    _addToolbar(model) {
        var _a, _b;
        const cell = this._getCell(model);
        if (cell) {
            const { leftMenu } = (_b = (_a = this._settings) === null || _a === void 0 ? void 0 : _a.composite) !== null && _b !== void 0 ? _b : {};
            const leftMenu_ = leftMenu === null ? [] : leftMenu !== null && leftMenu !== void 0 ? leftMenu : DEFAULT_LEFT_MENU;
            const toolbar = new CellToolbarWidget(this._commands, leftMenu_);
            toolbar.addClass(CELL_BAR_CLASS);
            cell.layout.insertWidget(0, toolbar);
            // For rendered markdown, watch for resize events.
            cell.displayChanged.connect(this._resizeEventCallback, this);
            // Watch for changes in the cell's contents.
            cell.model.contentChanged.connect(this._changedEventCallback, this);
        }
    }
    _getCell(model) {
        var _a;
        return (_a = this._panel) === null || _a === void 0 ? void 0 : _a.content.widgets.find(widget => widget.model === model);
    }
    _findToolbarWidgets(cell) {
        const widgets = cell.layout.widgets;
        // Search for header using the CSS class or use the first one if not found.
        return widgets.filter(widget => widget.hasClass(CELL_BAR_CLASS)) || [];
    }
    _removeToolbar(model) {
        const cell = this._getCell(model);
        if (cell) {
            this._findToolbarWidgets(cell).forEach(widget => widget.dispose());
            // Attempt to remove the resize and changed event handlers.
            cell.displayChanged.disconnect(this._resizeEventCallback, this);
            cell.model.contentChanged.disconnect(this._changedEventCallback, this);
        }
    }
    /**
     * Call back on settings changes
     */
    _onSettingsChanged() {
        var _a;
        // Reset toolbar when settings changes
        const activeCell = (_a = this._panel) === null || _a === void 0 ? void 0 : _a.content.activeCell;
        if (activeCell) {
            this._removeToolbar(activeCell.model);
            this._addToolbar(activeCell.model);
        }
    }
    _changedEventCallback() {
        var _a;
        const activeCell = (_a = this._panel) === null || _a === void 0 ? void 0 : _a.content.activeCell;
        if (activeCell === null || activeCell === undefined) {
            return;
        }
        this._updateCellForToolbarOverlap(activeCell);
    }
    _resizeEventCallback() {
        var _a;
        const activeCell = (_a = this._panel) === null || _a === void 0 ? void 0 : _a.content.activeCell;
        if (activeCell === null || activeCell === undefined) {
            return;
        }
        this._updateCellForToolbarOverlap(activeCell);
    }
    _updateCellForToolbarOverlap(activeCell) {
        // Remove the "toolbar overlap" class from the cell, rendering the cell's toolbar
        const activeCellElement = activeCell.node;
        activeCellElement.classList.remove(TOOLBAR_OVERLAP_CLASS);
        if (this._cellToolbarOverlapsContents(activeCell)) {
            // Add the "toolbar overlap" class to the cell, completely concealing the toolbar,
            // if the first line of the content overlaps with it at all
            activeCellElement.classList.add(TOOLBAR_OVERLAP_CLASS);
        }
    }
    _cellToolbarOverlapsContents(activeCell) {
        const cellType = activeCell.model.type;
        // If the toolbar is too large for the current cell, hide it.
        const cellLeft = this._cellEditorWidgetLeft(activeCell);
        const cellRight = this._cellEditorWidgetRight(activeCell);
        const toolbarLeft = this._cellToolbarLeft(activeCell);
        if (toolbarLeft === null) {
            return false;
        }
        // The toolbar should not take up more than 50% of the cell.
        if ((cellLeft + cellRight) / 2 > toolbarLeft) {
            return true;
        }
        if (cellType === 'markdown' && activeCell.rendered) {
            // Check for overlap in rendered markdown content
            return this._markdownOverlapsToolbar(activeCell);
        }
        // Check for overlap in code content
        return this._codeOverlapsToolbar(activeCell);
    }
    /**
     * Check for overlap between rendered Markdown and the cell toolbar
     *
     * @param activeCell A rendered MarkdownCell
     * @returns `true` if the first line of the output overlaps with the cell toolbar, `false` otherwise
     */
    _markdownOverlapsToolbar(activeCell) {
        const markdownOutput = activeCell.inputArea; // Rendered markdown appears in the input area
        // Get the rendered markdown as a widget.
        const markdownOutputWidget = markdownOutput.renderedInput;
        const markdownOutputElement = markdownOutputWidget.node;
        const firstOutputElementChild = markdownOutputElement.firstElementChild;
        if (firstOutputElementChild === null) {
            return false;
        }
        // Temporarily set the element's max width so that the bounding client rectangle only encompasses the content.
        const oldMaxWidth = firstOutputElementChild.style.maxWidth;
        firstOutputElementChild.style.maxWidth = 'max-content';
        const lineRight = firstOutputElementChild.getBoundingClientRect().right;
        // Reinstate the old max width.
        firstOutputElementChild.style.maxWidth = oldMaxWidth;
        const toolbarLeft = this._cellToolbarLeft(activeCell);
        return toolbarLeft === null ? false : lineRight > toolbarLeft;
    }
    _codeOverlapsToolbar(activeCell) {
        const editorWidget = activeCell.editorWidget;
        const editor = activeCell.editor;
        if (editor.lineCount < 1) {
            return false; // Nothing in the editor
        }
        const codeMirrorLines = editorWidget.node.getElementsByClassName('CodeMirror-line');
        if (codeMirrorLines.length < 1) {
            return false; // No lines present
        }
        const lineRight = codeMirrorLines[0].children[0] // First span under first pre
            .getBoundingClientRect().right;
        const toolbarLeft = this._cellToolbarLeft(activeCell);
        return toolbarLeft === null ? false : lineRight > toolbarLeft;
    }
    _cellEditorWidgetLeft(activeCell) {
        return activeCell.editorWidget.node.getBoundingClientRect().left;
    }
    _cellEditorWidgetRight(activeCell) {
        return activeCell.editorWidget.node.getBoundingClientRect().right;
    }
    _cellToolbarLeft(activeCell) {
        const toolbarWidgets = this._findToolbarWidgets(activeCell);
        if (toolbarWidgets.length < 1) {
            return null;
        }
        const activeCellToolbar = toolbarWidgets[0].node;
        return activeCellToolbar.getBoundingClientRect().left;
    }
}
/**
 * Widget extension that creates a CellToolbarTracker each time a notebook is
 * created.
 */
export class CellBarExtension {
    constructor(commands, settings) {
        this._commands = commands;
        this._settings = settings;
    }
    createNew(panel) {
        return new CellToolbarTracker(panel, this._commands, this._settings);
    }
}
//# sourceMappingURL=celltoolbartracker.js.map